from __future__ import annotations

import asyncio
import functools
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Generic, Type, cast

from pydantic import ValidationError

from .context import _current_message, _current_service
from .exceptions import DecodeError, Fail, Skip
from .logger import LoggerMixin
from .message import Message
from .middleware import Middleware
from .models import CloudEvent
from .settings import BrokerSettings
from .types import Encoder, RawMessage

if TYPE_CHECKING:
    from eventiq import Consumer, Service


class Broker(Generic[RawMessage], LoggerMixin, ABC):
    """Base broker class
    :param description: Broker (Server) Description
    :param encoder: Encoder (Serializer) class
    :param middlewares: Optional list of middlewares
    """

    protocol: str
    Settings = BrokerSettings

    def __init__(
        self,
        *,
        description: str | None = None,
        encoder: Encoder | None = None,
        middlewares: list[Middleware] | None = None,
    ) -> None:

        if encoder is None:
            from .encoders import get_default_encoder

            encoder = get_default_encoder()
        self.description = description or type(self).__doc__
        self.encoder = encoder
        self.middlewares: list[Middleware] = middlewares or []
        self._lock = asyncio.Lock()
        self._stopped = True

    def __repr__(self):
        return type(self).__name__

    @property
    def message_proxy_class(self) -> type[Message]:
        return Message

    def get_handler(
        self, service: Service, consumer: Consumer
    ) -> Callable[[RawMessage], Awaitable[Any | None]]:
        async def handler(raw_message: RawMessage) -> None:
            exc: Exception | None = None
            result: Any = None
            msg = self.message_proxy_class(raw_message)
            s_token = _current_service.set(service)
            try:
                parsed = self.parse_incoming_message(raw_message)
                message = consumer.validate_message(parsed)
                message.set_raw(msg)
            except (DecodeError, ValidationError) as e:
                self.logger.exception("Message parsing error", exc_info=e)
                msg.fail()
                await self.ack(service, consumer, msg)
                return
            m_token = _current_message.set(message)
            try:
                await self.dispatch_before(
                    "process_message", service, consumer, message
                )
            except Skip:
                self.logger.info(f"Skipped message {message.id}")
                await self.dispatch_after("skip_message", service, consumer, message)
                await self.ack(service, consumer, msg)
                return
            try:
                self.logger.info(
                    f"Running consumer {consumer.name} with message {message.id}"
                )
                result = await asyncio.wait_for(
                    consumer.process(message), timeout=consumer.timeout
                )
                if consumer.forward_response and result is not None:
                    await self.publish_event(
                        CloudEvent(
                            type=consumer.forward_response.as_type,
                            topic=consumer.forward_response.topic,
                            data=result,
                            source=service.name,
                        )
                    )
            # TODO: asyncio.CanceledError handling (?)
            except Exception as e:
                self.logger.error(f"Exception in {consumer.name} {e}")
                exc = e
            finally:
                self.logger.info(f"Finished running consumer {consumer.name}")
                await self.dispatch_after(
                    "process_message", service, consumer, message, result, exc
                )
                if exc is None or isinstance(exc, Fail) or msg.failed:
                    await self.ack(service, consumer, msg)
                else:
                    await self.nack(service, consumer, msg)
                _current_service.reset(s_token)
                _current_message.reset(m_token)

        return handler

    async def ack(self, service: Service, consumer: Consumer, message: Message) -> None:
        await self.dispatch_before("ack", service, consumer, message)
        await self._ack(message)
        await self.dispatch_after("ack", service, consumer, message)

    async def nack(
        self,
        service: Service,
        consumer: Consumer,
        message: Message,
    ) -> None:
        await self.dispatch_after(
            "nack",
            service,
            consumer,
            message,
        )
        await self._nack(message)
        await self.dispatch_after(
            "nack",
            service,
            consumer,
            message,
        )

    async def connect(self) -> None:
        async with self._lock:
            if self._stopped:
                await self.dispatch_before("broker_connect")
                await self._connect()
                self._stopped = False
                await self.dispatch_after("broker_connect")

    async def disconnect(self) -> None:
        async with self._lock:
            if not self._stopped:
                await self.dispatch_before("broker_disconnect")
                await self._disconnect()
                self._stopped = True
                await self.dispatch_after("broker_disconnect")

    async def publish_event(self, message: CloudEvent, **kwargs: Any) -> None:
        """
        :param message: Cloud event object to send
        :param kwargs: Additional params passed to broker._publish
        :rtype: None
        """
        await self.dispatch_before("publish", message)
        await self._publish(message, **kwargs)
        await self.dispatch_after("publish", message)

    async def publish(
        self,
        topic: str,
        data: Any | None = None,
        type_: type[CloudEvent] | str = "CloudEvent",
        source: str = "",
        **kwargs: Any,
    ) -> None:
        """Publish message to broker
        :param topic: Topic to publish data
        :param data: Message content
        :param type_: Event type, name or class
        :param source: message sender (service/app name)
        :param kwargs: additional params passed to underlying broker implementation, such as headers
        :rtype: None
        """

        if isinstance(type_, str):
            cls = functools.partial(CloudEvent, type=type_)
        else:
            cls = type_  # type: ignore

        message: CloudEvent = cls(
            content_type=self.encoder.CONTENT_TYPE,
            topic=topic,
            data=data,
            source=source,
        )
        await self.publish_event(message, **kwargs)

    async def start_consumer(self, service: Service, consumer: Consumer):
        await self.dispatch_before("consumer_start", service, consumer)
        await self._start_consumer(service, consumer)
        await self.dispatch_after("consumer_start", service, consumer)

    def add_middleware(self, middleware: Middleware | type[Middleware]) -> None:
        if isinstance(middleware, type):
            middleware = middleware()
        self.middlewares.append(middleware)

    def add_middlewares(self, middlewares: list[Middleware | type[Middleware]]) -> None:
        for m in middlewares:
            self.add_middleware(m)

    async def _dispatch(self, full_event: str, *args, **kwargs) -> None:
        for m in self.middlewares:
            try:
                await getattr(m, full_event)(self, *args, **kwargs)
            except Skip:
                raise
            except Exception as e:
                self.logger.exception("Unhandled middleware exception", exc_info=e)

    async def dispatch_before(self, event: str, *args, **kwargs) -> None:
        await self._dispatch(f"before_{event}", *args, **kwargs)

    async def dispatch_after(self, event: str, *args, **kwargs) -> None:
        await self._dispatch(f"after_{event}", *args, **kwargs)

    @classmethod
    def from_settings(cls, settings: BrokerSettings, **kwargs: Any) -> Broker:
        broker_class: type[Broker] = cast(Type[Broker], settings.broker_class)
        kw = settings.dict(exclude={"broker_class"})
        kw.update(kwargs)
        return broker_class(**kw)

    @classmethod
    def from_env(
        cls,
        **kwargs,
    ) -> Broker:
        return cls.from_settings(BrokerSettings(), **kwargs)

    @abstractmethod
    def parse_incoming_message(self, message: RawMessage) -> Any:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        """Return broker connection status"""
        raise NotImplementedError

    @abstractmethod
    async def _publish(self, message: CloudEvent, **kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _connect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _disconnect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        raise NotImplementedError

    async def _ack(self, message: Message) -> None:
        """Empty default implementation for backends that do not support explicit ack"""

    async def _nack(self, message: Message) -> None:
        """Same as for ._ack()"""



from datetime import datetime, timedelta
from typing import Any, Dict, Generic, Optional

from pydantic import Extra, Field, validator
from pydantic.fields import PrivateAttr
from pydantic.generics import GenericModel

from .context import get_current_service
from .message import Message
from .types import ID, D
from .utils import str_uuid
from .utils.datetime import utc_now


class CloudEvent(GenericModel, Generic[D]):
    specversion: Optional[str] = "1.0"
    content_type: str = Field("application/json", alias="datacontenttype")
    id: ID = Field(default_factory=str_uuid)
    time: datetime = Field(default_factory=utc_now)
    topic: str = Field(..., alias="subject")
    type: Optional[str] = None
    source: Optional[str] = None
    data: D
    trace_ctx: Dict[str, str] = {}

    _raw: Optional[Any] = PrivateAttr()

    def __eq__(self, other):
        if not isinstance(other, CloudEvent):
            return False
        return self.id == other.id

    @validator("type", allow_reuse=True, always=True, pre=True)
    def get_type_from_cls_name(cls, v) -> str:
        return v or cls.__name__

    @property
    def raw(self) -> Message:
        if self._raw is None:
            raise AttributeError("raw property accessible only for incoming messages")
        return self._raw

    @property
    def service(self):
        return get_current_service()

    @property
    def log_context(self) -> Dict[str, Any]:
        return {"message_id": str(self.id)}

    def _set(self, name: str, value: Any):
        object.__setattr__(self, name, value)

    def set_source(self, value: str) -> None:
        self._set("source", value)

    def set_raw(self, value: Message):
        self._set("_raw", value)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        kwargs.setdefault("by_alias", True)
        return super().dict(**kwargs)

    @classmethod
    def new(cls, obj: D, **kwargs: Any):
        return cls(data=obj, **kwargs)

    @property
    def age(self) -> timedelta:
        return utc_now() - self.time

    def fail(self):
        self.raw.fail()

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True
        extra = Extra.allow
        # events are immutable, however it's sometimes useful to set source or traceid for unpublished events
        allow_mutation = False



from __future__ import annotations

import functools
import json
from collections import defaultdict
from itertools import chain
from pathlib import Path

from pydantic.schema import schema as all_schemas

from eventiq.asyncapi.models import (
    AsyncAPI,
    ChannelItem,
    Components,
    Info,
    Message,
    Operation,
    PayloadRef,
)

from ..service import Service
from .registry import PUBLISH_REGISTRY

PREFIX = "#/components/schemas/"


def normalize(v: str) -> str:
    return v.lower().replace("-", "_")


def get_all_models_schema(service: Service):
    all_models = [
        m.event_type  # type: ignore
        for m in chain(service.consumers.values(), PUBLISH_REGISTRY.values())
    ]
    return all_schemas(all_models, ref_prefix=PREFIX).get("definitions", [])


def populate_channel_spec(service: Service):
    channels: dict[str, ChannelItem] = defaultdict(ChannelItem)
    # tags = {t["name"]: Tag(**t) for t in service.tags_metadata}

    for publishes in PUBLISH_REGISTRY.values():
        channels[publishes.topic].publish = Operation(
            message=Message(
                message_id=normalize(f"{service.name}_{publishes.event_type.__name__}"),
                content_type=service.broker.encoder.CONTENT_TYPE,
                payload=PayloadRef(ref=f"{PREFIX}{publishes.event_type.__name__}"),
                description=publishes.kwargs.get("description", ""),
            )
        )
    for consumer in service.consumers.values():

        subscribe = Operation(
            message=Message(
                message_id=normalize(f"{consumer.name}_{consumer.event_type.__name__}"),
                content_type=service.broker.encoder.CONTENT_TYPE,
                payload=PayloadRef(ref=f"{PREFIX}{consumer.event_type.__name__}"),
                description=consumer.description,
            )
        )
        channels[consumer.topic].subscribe = subscribe
    return channels


@functools.lru_cache
def get_async_api_spec(service: Service) -> AsyncAPI:
    channels = populate_channel_spec(service)
    definitions = get_all_models_schema(service)
    doc_model = AsyncAPI(
        info=Info(title=service.title, version=service.version),
        servers={},
        channels=channels,
        components=Components(schemas=definitions),
        default_content_type=service.broker.encoder.CONTENT_TYPE,
    )
    return doc_model


def save_async_api_to_file(spec: AsyncAPI, path: Path, fmt: str) -> None:
    dump = json.dump
    if fmt == "yaml":
        import yaml

        dump = yaml.dump
    with open(path, "w") as f:
        dump(spec.dict(by_alias=True, exclude_none=True), f)
